# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H.Payno", "J.Garriga"]
__license__ = "MIT"
__date__ = "14/10/2019"

import os
import tempfile

import fabio
import numpy
import string

from darfix.core.dataset import Dataset


def random_generator(size=4, chars=string.printable):
    """
    Returns a string with random characters.
    """
    return ''.join(chars[numpy.random.choice(len(chars))] for x in range(size))


def createRandomDataset(dims, nb_data_files=10, nb_dark_files=2, header=False, _dir=None):
    """Simple creation of a dataset in _dir with the requested number of data
    files and dark files.

    :param tuple of int dims: dimensions of the files.
    :param int nb_data_files: Number of data files to create.
    :param int nb_data_files: Number of dark files to create.
    :param bool header: If True, a random header is created for every frame.
    :param str or None _dir: Directory to save the temporary files.

    :return :class:`Dataset`: generated instance of :class:`Dataset`
    """
    assert type(dims) is tuple and len(dims) == 2
    assert type(nb_data_files) is int
    assert type(nb_dark_files) is int
    # assert type(nb_ff_file) is int
    assert type(_dir) in (type(None), str)
    assert nb_dark_files > 0

    dark_filename = None

    if _dir is None:
        _dir = tempfile.mkdtemp()

    if os.path.isdir(_dir) is False:
        raise ValueError("%s is not a directory" % _dir)

    # Number of counters and motors for the dataset frames.
    counter_values = numpy.random.randint(50)
    motor_values = numpy.random.randint(50)
    # Creating the random strings to compose the counters and motors names.
    counter_mne = [random_generator() for i in numpy.arange(counter_values)]
    motor_mne = [random_generator() for i in numpy.arange(motor_values)]
    for index in range(nb_data_files):
        header_dict = {}
        if header:
            # Create random header
            header_dict["HeaderID"] = index
            header_dict["Image"] = 1
            header_dict["counter_pos"] = numpy.random.random(counter_values)
            header_dict["counter_mne"] = counter_mne
            header_dict["motor_pos"] = numpy.random.random(motor_values)
            header_dict["motor_mne"] = motor_mne
        data_file = os.path.join(_dir, 'data_file%04i.edf' % index)
        image = fabio.edfimage.EdfImage(data=numpy.random.random(dims), header=header_dict)
        image.write(data_file)
    raw_filename = os.path.join(_dir, 'data_file%04i.edf' % 0)
    for index in range(nb_dark_files):
        dark_file = os.path.join(_dir, 'dark_file%04i.edf' % index)
        image = fabio.edfimage.EdfImage(data=numpy.random.random(dims))
        image.write(dark_file)
    if nb_dark_files > 0:
        dark_filename = os.path.join(_dir, 'dark_file%04i.edf' % 0)

    dataset = Dataset(raw_filename=raw_filename, dark_filename=dark_filename)
    return dataset


def createDataset(data, dark_frames=None, filter_data=False, header=None, _dir=None):
    """
    Create a dataset from a configuration

    :param numpy.ndarray data: Images to form the data.
    :param numpy.ndarray dark_frames: Images to form the dark frames.
    :param bool filter_data: If True, the dataset created will divide the data
        between the ones with no intensity (or very low) and the others.
    :param Union[None,array_like] header: List with a header per frame. If None,
        no header is added.
    :param str or None _dir: Directory to save the temporary files.

    :return :class:`Dataset`: generated instance of :class:`Dataset`.
    """
    assert type(_dir) in (type(None), str)
    assert len(data) > 0
    if header is not None:
        assert len(header) == len(data)

    if _dir is None:
        _dir = tempfile.mkdtemp()

    if os.path.isdir(_dir) is False:
        raise ValueError("%s is not a directory" % _dir)
    for index in range(len(data)):
        data_file = os.path.join(_dir, 'data_file%04i.edf' % index)
        if header is not None:
            image = fabio.edfimage.EdfImage(data=data[index], header=header[index])
        else:
            image = fabio.edfimage.EdfImage(data=data[index])

        image.write(data_file)
    raw_filename = os.path.join(_dir, 'data_file%04i.edf' % 0)

    dark_filename = None
    if dark_frames is not None:
        for index in range(len(dark_frames)):
            data_file = os.path.join(_dir, 'dark_file%04i.edf' % index)
            image = fabio.edfimage.EdfImage(data=dark_frames[index])
            image.write(data_file)
        dark_filename = os.path.join(_dir, 'dark_file%04i.edf' % 0)

    dataset = Dataset(raw_filename=raw_filename, dark_filename=dark_filename,
                      filter_data=filter_data)

    return dataset
